/*
Lexmark 1100 printer driver for Linux
Copyright (C) 1999   Tim Engler

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "LexmarkProtocol300c.h"
#include "PaintLine300c.h"
#include "errors.h"
#include <stdio.h>
#include <stdarg.h>
#include "Jet.h"

// a very stateful implementation of protocol. This class actually sends codes to printer.

#define PAGE_Y_START 85 //initial y position of page
#define PRINT_Y_START 100 //initial y position of print zone(this will offset all coordinates given to paint)
#define TOP_BORDER PRINT_Y_START - PAGE_Y_START
#define BOTTOM_BORDER 140 //y position from the bottom of the page which we will refuse to print at
#define RIGHT_BORDER 216 //x position from right of the page which we will refuse to print at
#define PRINT_X_START 1141 //initial x position of print zone(this will offset all coordinates given to paint)
#define PAGE_X_START 1092 //initial x position of page
#define LEFT_BORDER PRINT_X_START - PAGE_X_START

LexmarkProtocol300c::LexmarkProtocol300c(Writer *writer)
{
	paintLine = 0;
	lastPaintLine = 0;
	y = 0;
	firstPaintLine = true;
	pageStarted = false;

	this->writer = writer;
}

LexmarkProtocol300c::~LexmarkProtocol300c()
{
	if(paintLine)
		delete paintLine;
	if(lastPaintLine)
		delete lastPaintLine;
}

Rect LexmarkProtocol300c::getPageSize(int pageType)
{
	if(pageType == LETTER_PAGE)
	//300x300 dpi 8 1/2 x 11 inch page
		return Rect(0,0,2550,3300);

	die("Bad page type %d",pageType);
	return Rect();
}


//returns the total number of jets, creates new jets that must be freed by the client
Jet **LexmarkProtocol300c::createJets(int *numJets, Map *imageMap)
{
	*numJets = 6;
	class Jet **jets = new Jet*[6];
	jets[0] = new Jet(0,9,0,imageMap,LEFT_JET,0,8);
	jets[1] = new Jet(1,0,0,imageMap,RIGHT_JET,0,8);
	jets[2] = new Jet(2,-9,20,imageMap,LEFT_JET,1,8);
	jets[3] = new Jet(3,-18,20,imageMap,RIGHT_JET,1,8);
	jets[4] = new Jet(4,8,40,imageMap,LEFT_JET,2,8);
	jets[5] = new Jet(5,1,40,imageMap,RIGHT_JET,2,8);	

	return jets;
}


void LexmarkProtocol300c::startPaint(int startX, int startY, int size)
{
    if(!pageStarted)
		die("Page never started, call start_page");

    if(lastPaintLine)
    {
		printPaintLine(lastPaintLine, paintLine->x);
		delete lastPaintLine;
		lastPaintLine = 0;
    }

    if(paintLine)
    {
		lastPaintLine = paintLine;
    }

    paintLine = new PaintLine300c(startX, startY, size);
}

//private method to advance the page to an absolute position
//if last, the page is ejected
void LexmarkProtocol300c::advancePage(int y,bool last)
{
    y += PAGE_Y_START;

    if(y < this->y)
		die("Negative movement asked for in advance_page");

	if(last)
		printData(2,0x22,0x80);
	else
		printData(2,0x23,0x80);
    printInt((y-this->y) * 2);
    printData(4,0x00,0x00,0x00,0x00);

    this->y = y;
}

//private method to print out the paint line
void LexmarkProtocol300c::printPaintLine(PaintLine300c *currentPaintLine, int nextLineStart)
{
    //advance page
    advancePage(currentPaintLine->y,false);

    // send the paint packet
    printData(2,0x40,0x84);

    int dataPackets = currentPaintLine->size;
    int printerStart = (currentPaintLine->x + PAGE_X_START) * 2;
    
    printInt(dataPackets);
    
    //proportional to length of data
    if(firstPaintLine)
    {
		printInt((int)(dataPackets/2. + (printerStart)/4. - 373.5));
		firstPaintLine = false;
    }
    else
    {
		printInt(dataPackets/2 + 182);
    }
    
    //printer start
    printInt(printerStart);
    
    if(nextLineStart != -1)
    {
		//always is the same except last paint
		printData(2,0x41,0x10);

		//this should equal the end of this line to the start of next line
	
		if(currentPaintLine->size + currentPaintLine->x < nextLineStart) 
			//if negative we just use 0. We don't want the head to be smacking into the
			//edge of the printer if it reads the data as unsigned, or who knows what else?
		{
			//fixme - we should really find out what the driver should do in this case
			printInt(58);
		}
		else 
		{
			printInt((currentPaintLine->size + currentPaintLine->x - nextLineStart+1)/2  + 58);
		}
    }
    else
    {
		//last paint has this
		printData(2,0x41,0x00);
		printData(2,0x00,0x00);
    }
    
    printData(7,0x00,0x00,0x00,0x00,0x80,0x00,0x00);
    
    //bytes to print
    printInt(dataPackets * 6);
    
    printData(3,0x00,0x00,0x00);
    
	for(int i = 0; i < currentPaintLine->size*2; i+=2)
	{
		for(int j = 0; j < 3; j++)
		{
			writer->write(&currentPaintLine->data[j][i],1);
			writer->write(&currentPaintLine->data[j][i+1],1);
		}
	}
}

void LexmarkProtocol300c::paint(int sx,int x,int nozzle,int color)
{
    paintLine->paint(sx,x, nozzle, color);
}

void LexmarkProtocol300c::endPaint()
{
}

void LexmarkProtocol300c::endPage()
{
    pageStarted = false;

    if(lastPaintLine)
    {
		printPaintLine(lastPaintLine, paintLine->x);
    }

    if(paintLine)
    {
		//last paint 
		printPaintLine(paintLine,-1);
    }

	//finish page
    printData(16,0x31,0x18,0x00,0xb6,0x00,0x00,0x00,0x00,0x33,0x00,0x00,0x00,0x00,0x00,0x00,0x00);

	advancePage(3471,true);
}

int LexmarkProtocol300c::getLeftBorder()
{
	return LEFT_BORDER;
}


int LexmarkProtocol300c::getTopBorder()
{
	return TOP_BORDER;
}

int LexmarkProtocol300c::getBottomBorder()
{
	return BOTTOM_BORDER;
}


int LexmarkProtocol300c::getRightBorder()
{
	return RIGHT_BORDER;
}

//adjust the starting x and y positions for the granualarity of the protocol and print zone
void LexmarkProtocol300c::adjustPositions(int *startX, int *y, int *endX)
{
	//x must be divisible by 4

    *startX -= (*startX % 2+1);
	*endX += 4 - (*endX - *startX) % 4;
}



//starts the page!
void LexmarkProtocol300c::startPage()
{
    pageStarted = true;
    printData(112,0x1b,0x2a,0x6d,0x00,0x00,0x00,0x00,0x00,0x20,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x21,0x00,0x00,0x00,
			  0x00,0x00,0x00,0x00,0x33,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x30,0xc0,0x40,0x01,0x00,0x00,0x00,0x00,
			  0x21,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x21,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x21,0x00,0x00,0x00,
			  0x00,0x00,0x00,0x00,0x21,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x21,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
			  0x21,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x21,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x21,0x00,0x00,0x00,
			  0x00,0x00,0x00,0x00,0x21,0x00,0x00,0x00,0x00,0x00,0x00,0x00);
}


//private method for printing data
void LexmarkProtocol300c::printData(int len, ...)
{
	va_list ap;
	va_start(ap,len);

    for(int i = 0; i < len; i++)
    {
		char c = (char)va_arg(ap,int);
		
		writer->write(&c,1);
    }
}

//private method for printing an int
void LexmarkProtocol300c::printInt(int i)
{
    //printer data is big endian
    printData(2,i >> 8, i % 256);
}


