/*
Lexmark 1100 printer driver for Linux
Copyright (C) 1999   Tim Engler

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "ThreadSafeCircleBuffer.h"

ThreadSafeCircleBuffer::ThreadSafeCircleBuffer(int maxSize)
{
	this->maxSize = maxSize;
	this->buffer = new char[maxSize];
	this->pos = 0;
	this->size = 0;

	this->consumerSleeping = false;
	this->producerSleeping = false;
	
	pthread_mutex_init(&this->mutex,NULL);
	pthread_cond_init(&this->cond,NULL);

	this->valid = true;
}


// This object should not be deleted while another thread is writing to
// it. It can be deleted before it is emptied however. The thread
// will be suspended until the buffer is emptied by the consumer
// thread
ThreadSafeCircleBuffer::~ThreadSafeCircleBuffer()
{

	// we lock here so the consumer thread doesn't try to wake
	// us up while we are still awake, since we are setting
	// valid to false and waking it up before we go to sleep 
	pthread_mutex_lock(&mutex);
	
	valid = false;

	//if there is stuff left in the buffer, we sleep
	//so the consumer can clean it out

	// if size is not zero, the consumer will never be
	// sleeping so we can be assurred it will wake us up
	// It knows to do this since valid is false
	if(size != 0) 
	{
		//		fprintf(stderr,"destructor sleeping\n");
		pthread_cond_wait(&cond,&mutex);
	}

	// if we were sleeping
	if(consumerSleeping)
	{
		//		fprintf(stderr,"destructor waking up get thread\n");
		
		pthread_cond_signal(&cond); //wake up, it's time to go home
		pthread_mutex_unlock(&mutex);
		return;
	}
	
	pthread_mutex_unlock(&mutex);

	delete buffer;
}


// adds data to the buffer, if the buffer becomes full, the
// thread is suspended until it starts to empty
void ThreadSafeCircleBuffer::put(char *data, int dataLength)
{
	do
	{
		// no more room
		if(maxSize == size)
		{
			//I ate too much..I'm sleepy
			pthread_mutex_lock(&mutex);
			producerSleeping = true;
			pthread_cond_wait(&cond,&mutex);
			pthread_mutex_unlock(&mutex);
		}
		
		int fillLength;

		//we lock here since we are reading the size and the pos, which may change
		//at anytime , even between the time we read pos, and then read size
		pthread_mutex_lock(&mutex);
		int fillStart = (pos + size) % maxSize;
		pthread_mutex_unlock(&mutex);
		
		// either there's some room left, just not all that we would like
		// or the pos isn't near enough the beginning of the circle
		if(dataLength > maxSize - fillStart)
		{
			//fill up what we can now and try again
			//the pos will be moved to the start, if there is some room left,
			//or we'll wait for more data in top if
			fillLength = maxSize - fillStart;
		}
		else fillLength = dataLength;
		
		// write in the data
		strncpy(
&buffer[fillStart],
(char *)data,fillLength);

		//we do not lock here since even though we are changing size, we will
		//only increase it. 
		//As long as get() gets a size which equal or less than the true size
		//it will be ok 
		size += fillLength;
		
		// adjust data pointer for fresh data
		data += fillLength;
		dataLength -= fillLength;

		//we must lock here so the get() thread doesn't mark itself sleeping,
		//give up control, and then we think it is sleeping and wake it up
		//although it hasn't gone to sleep yet. And then it would sleep even
		//though there was data to read
		pthread_mutex_lock(&mutex);
		if(consumerSleeping)
		{
			//			fprintf(stderr,"put thread waking up get thread\n");
			pthread_cond_signal(&cond);

			//we set the consumer as not sleeping here, incase it doesn't get a lock 
			//by the time we wrap around again or the destructor is called.
			consumerSleeping = false;
		}
		
		pthread_mutex_unlock(&mutex);
		
	}
	while(dataLength > 0); // while there is data left to transmit
}
	
// pulls data out of the buffer and sends it to an output stream. If there is no data,
// thread blocks until there is or the buffer is destroyed
void ThreadSafeCircleBuffer::get(int fd)
{
	pthread_mutex_lock(&mutex);

	while(1)
	{
		
		//if there is nothing to send
		if(size == 0)
		{
			if(!valid) {// if the buffer is shutting down, we quit, no more data
				pthread_cond_signal(&cond);
				pthread_mutex_unlock(&mutex);
				return;
			}

			//otherwise we sleep 
			consumerSleeping = true;
			//			fprintf(stderr,"get thread sleeping\n");
			
			pthread_cond_wait(&cond,&mutex);
			//			fprintf(stderr,"get thread woke up\n");
			
			
			// even if we are destroying the object, we have to wait for it
			// to be emptied to be a good buffer, so we don't exit.
		}
		pthread_mutex_unlock(&mutex);
		
		int drainLength = maxSize - pos;
		
		if(drainLength > size) // we don't want to send more than we have
		drainLength = size;
		
		// write the data
		// we don't lock here because if put() gets control it is ok, since
		// the available room is only becoming greater.
		write(fd, &buffer[pos], drainLength);
		
		// we lock here since we are changing the pos and the size, and we
		// don't want put() to catch the size and the pos in inconsistant states
		pthread_mutex_lock(&mutex);
		pos = (pos + drainLength) % maxSize;
		size -= drainLength;
		
		//we must lock here so the put() thread doesn't mark itself sleeping,
		//give up control, and then we think it is sleeping and wake it up
		//although it hasn't gone to sleep yet. And then it would sleep even
		//though there was room to write
		if(producerSleeping)
		{
			pthread_cond_signal(&cond);

			//we set producer as not sleeping here, incase it doesn't get a lock 
			//by the time we wrap around again.
			producerSleeping = false;
		}
	}
}


bool ThreadSafeCircleBuffer::hasData()
{
	return size != 0;
}
